#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import time
import typing

import pytest
from ccxt import Exchange

import octobot_commons.constants as constants
import octobot_commons.enums as commons_enums
import octobot_trading.enums as trading_enums
import octobot_trading.errors as trading_errors
import octobot_trading.exchanges.connectors.ccxt.ccxt_client_util as ccxt_client_util
import octobot_trading.exchanges.util as exchanges_util
from octobot_trading.enums import ExchangeConstantsTickersColumns as Ectc, \
    ExchangeConstantsMarketStatusColumns as Ecmsc
from tests_additional.real_exchanges import get_exchange_manager


class RealExchangeTester:
    # enter exchange name as a class variable here
    EXCHANGE_NAME = None
    EXCHANGE_TYPE = trading_enums.ExchangeTypes.SPOT.value
    SYMBOL = None
    SYMBOL_2 = None
    SYMBOL_3 = None
    INACTIVE_MARKETS = []
    # default is 1h, change if necessary
    TIME_FRAME = commons_enums.TimeFrames.ONE_HOUR
    ALLOWED_TIMEFRAMES_WITHOUT_CANDLE = 0
    CANDLE_SINCE = 1661990400000  # 1 September 2022 00:00:00
    CANDLE_SINCE_SEC = CANDLE_SINCE / 1000
    REQUIRES_AUTH = False  # set True when even normally public apis require authentication
    MARKET_STATUS_TYPE = trading_enums.ExchangeTypes.SPOT.value
    HISTORICAL_CANDLES_TO_FETCH_COUNT = 650
    DEFAULT_CUSTOM_BOOK_LIMIT = 50
    ORDER_BOOK_DESYNC_ALLOWANCE = 60    # allow 60s desync

    # Public methods: to be implemented as tests
    # Use await self._[method_name] to get the test request result
    # ex: market_status = await self.get_market_status()

    # unauthenticated API
    async def test_time_frames(self):
        pass

    async def test_get_market_status(self):
        pass

    async def test_get_symbol_prices(self):
        pass

    async def test_get_historical_symbol_prices(self):
        pass

    async def test_get_kline_price(self):
        pass

    async def test_get_order_book(self):
        pass

    async def test_get_order_books(self):
        # implement if necessary
        pass

    async def test_get_recent_trades(self):
        pass

    async def test_get_price_ticker(self):
        pass

    async def test_get_all_currencies_price_ticker(self):
        pass

    # authenticated API
    # TODO
    # async def test_get_balance(self):
    #     pass
    #
    # async def test_get_order(self):
    #     pass
    #
    # async def test_get_all_orders(self):
    #     pass
    #
    # async def test_get_open_orders(self):
    #     pass
    #
    # async def test_get_closed_orders(self):
    #     pass
    #
    # async def test_get_my_recent_trades(self):
    #     pass
    #
    # async def test_cancel_order(self):
    #     pass
    #
    # async def test_create_order(self):
    #     pass

    def ensure_required_market_status_values(self, market_status):
            assert market_status
            assert market_status[Ecmsc.TYPE.value] == self.MARKET_STATUS_TYPE
            assert market_status[Ecmsc.SYMBOL.value] in (self.SYMBOL, self.SYMBOL_2, self.SYMBOL_3)
            assert market_status[Ecmsc.ACTIVE.value] \
                   is False if market_status[Ecmsc.SYMBOL.value] in self.INACTIVE_MARKETS else True
            assert market_status[Ecmsc.PRECISION.value]

    async def test_get_historical_ohlcv(self):
        # common implementation, should always work if candles history is supported
        historical_ohlcv = await self.get_historical_ohlcv()
        assert len(historical_ohlcv) > 500  # should be around 650
        self.ensure_elements_order(historical_ohlcv, commons_enums.PriceIndexes.IND_PRICE_TIME.value)
        self.ensure_unique_elements(historical_ohlcv, commons_enums.PriceIndexes.IND_PRICE_TIME.value)
        start, end = self.get_historical_ohlcv_start_and_end_times()
        max_candle_time = self.get_time_after_time_frames(start, len(historical_ohlcv))
        assert max_candle_time <= end
        assert max_candle_time <= self.get_time()
        # on some exchanges, a lot of candles are missing, ensure more than 1 fetch succeeded
        assert (
            self.HISTORICAL_CANDLES_TO_FETCH_COUNT * 0.85
            < len(historical_ohlcv)
            <= self.HISTORICAL_CANDLES_TO_FETCH_COUNT
        )
        for candle in historical_ohlcv:
            assert start <= candle[commons_enums.PriceIndexes.IND_PRICE_TIME.value] <= end

    def get_config(self):
        return {
            constants.CONFIG_EXCHANGES: {
                self.EXCHANGE_NAME: {
                    constants.CONFIG_EXCHANGE_TYPE: self.EXCHANGE_TYPE
                }
            }
        }

    def _get_ref_order_book_timestamp(self):
        return time.time() + self.ORDER_BOOK_DESYNC_ALLOWANCE

    async def inner_test_unsupported_get_order_books(self):
        async with (self.get_exchange_manager() as exchange_manager):
            with pytest.raises(trading_errors.NotSupported):
                # with symbols filter
                symbols = [self.SYMBOL, self.SYMBOL_2, self.SYMBOL_3]
                await exchange_manager.exchange.get_order_books(symbols=symbols)

            with pytest.raises(trading_errors.NotSupported):
                # without symbols filter
                await exchange_manager.exchange.get_order_books(symbols=None)

    async def inner_test_get_order_books(
        self,
        expected_symbol_filter: bool,
        expected_missing_symbol_filter_books_min_count: int,
        expected_max_orders_by_side: int,
        min_book_orders_count: int,
        supports_custom_limit: bool,
        custom_limit: typing.Optional[int],
        max_empty_book_sides: int,
    ):
        async with (self.get_exchange_manager() as exchange_manager):
            # with symbols filter
            symbols = [self.SYMBOL, self.SYMBOL_2, self.SYMBOL_3]
            books_by_symbol = await exchange_manager.exchange.get_order_books(symbols=symbols)
            ref_time = self._get_ref_order_book_timestamp()
            empty_book_sides = []
            if expected_symbol_filter:
                assert len(books_by_symbol) == len(symbols)
            else:
                assert len(books_by_symbol) >= expected_missing_symbol_filter_books_min_count, (
                    f"{len(books_by_symbol)} < {expected_missing_symbol_filter_books_min_count}"
                )
            for symbol, book in books_by_symbol.items():
                assert 0 < book[trading_enums.ExchangeConstantsOrderBookInfoColumns.TIMESTAMP.value] < ref_time, (
                    f"Invalid {symbol} book timestamp value: "
                    f"{book[trading_enums.ExchangeConstantsOrderBookInfoColumns.TIMESTAMP.value]}, {ref_time=}"
                )
                for side in [
                    trading_enums.ExchangeConstantsOrderBookInfoColumns.BIDS,
                    trading_enums.ExchangeConstantsOrderBookInfoColumns.ASKS
                ]:
                    if len(book[side.value]) == 0:
                        empty_book_sides.append((symbol, side))
                    else:
                        assert min_book_orders_count <= len(book[side.value]) <= expected_max_orders_by_side, (
                            f"Unexpected {symbol} {side.value} orders count: {len(book[side.value])}. Expected: "
                            f"[{min_book_orders_count}:{expected_max_orders_by_side}]"
                        )
            if len(empty_book_sides) > max_empty_book_sides:
                raise AssertionError(
                    f"More empty book sides than expected: {max_empty_book_sides=} {empty_book_sides=}"
                )
            # without symbols filter
            books_by_symbol = await exchange_manager.exchange.get_order_books(symbols=None)
            empty_book_sides = []
            assert len(books_by_symbol) >= expected_missing_symbol_filter_books_min_count, (
                f"{len(books_by_symbol)=} NOT >= {expected_missing_symbol_filter_books_min_count=}"
            )
            for symbol, book in books_by_symbol.items():
                assert 0 < book[trading_enums.ExchangeConstantsOrderBookInfoColumns.TIMESTAMP.value] < ref_time, (
                    f"Invalid {symbol} book timestamp value: "
                    f"{book[trading_enums.ExchangeConstantsOrderBookInfoColumns.TIMESTAMP.value]}, {ref_time=}"
                )
                for side in [
                    trading_enums.ExchangeConstantsOrderBookInfoColumns.BIDS,
                    trading_enums.ExchangeConstantsOrderBookInfoColumns.ASKS
                ]:
                    if len(book[side.value]) == 0:
                        empty_book_sides.append((symbol, side))
                    else:
                        assert min_book_orders_count <= len(book[side.value]) <= expected_max_orders_by_side, (
                            f"Unexpected {symbol} {side.value} orders count: {len(book[side.value])}. Expected: "
                            f"[{min_book_orders_count}:{expected_max_orders_by_side}]"
                        )
            if len(empty_book_sides) > max_empty_book_sides:
                raise AssertionError(
                    f"More empty book sides than expected: {max_empty_book_sides=} {len(empty_book_sides)=}"
                )
            # with custom limit
            books_by_symbol = await exchange_manager.exchange.get_order_books(symbols=None, limit=custom_limit)
            self._ensure_book_custom_limit(
                books_by_symbol, supports_custom_limit, expected_max_orders_by_side, custom_limit
            )

    def _ensure_book_custom_limit(
        self,
        books_by_symbol: dict,
        supports_custom_limit: bool,
        no_limit_default_size: int,
        custom_limit: typing.Optional[int],
    ):
        has_taken_limit_into_account = False
        custom_limit = custom_limit or self.DEFAULT_CUSTOM_BOOK_LIMIT
        for symbol, book in books_by_symbol.items():
            for side in [
                trading_enums.ExchangeConstantsOrderBookInfoColumns.BIDS,
                trading_enums.ExchangeConstantsOrderBookInfoColumns.ASKS
            ]:
                if no_limit_default_size < len(book[side.value]) <= custom_limit:
                    has_taken_limit_into_account = True
        if supports_custom_limit:
            assert has_taken_limit_into_account
        else:
            assert not supports_custom_limit

    @contextlib.asynccontextmanager
    async def get_exchange_manager(self, market_filter=None):
        async with get_exchange_manager(
            self.EXCHANGE_NAME, config=self.get_config(),
            authenticated=self.REQUIRES_AUTH, market_filter=market_filter
        ) as exchange_manager:
            yield exchange_manager

    async def time_frames(self):
        async with self.get_exchange_manager() as exchange_manager:
            return exchange_manager.exchange.time_frames

    async def test_active_symbols(self):
        raise NotImplementedError("test_active_symbols is not implemented")

    async def inner_test_active_symbols(self, expected_active_symbols_count: int, expected_total_symbols_count: int):
        async with self.get_exchange_manager() as exchange_manager:
            # ensure active symbols are correctly parsed by ccxt
            active_symbols = exchange_manager.exchange.get_all_available_symbols(active_only=True)
            assert expected_active_symbols_count <= len(active_symbols) <= expected_active_symbols_count * 1.5, \
                f"{len(active_symbols)=}, {expected_active_symbols_count=}"
            all_symbols = exchange_manager.exchange.get_all_available_symbols(active_only=False)
            assert expected_total_symbols_count <= len(all_symbols) <= expected_total_symbols_count * 1.5, \
                f"{len(all_symbols)=}, {expected_total_symbols_count=}"

    async def get_market_statuses(self):
        # return 2 different market status with different traded pairs to reduce possible
        # side effects using only one pair.
        async with self.get_exchange_manager() as exchange_manager:
            self._ensure_market_status_cachability(exchange_manager)
            return exchange_manager.exchange.get_market_status(self.SYMBOL), \
                exchange_manager.exchange.get_market_status(self.SYMBOL_2), \
                exchange_manager.exchange.get_market_status(self.SYMBOL_3)

    def _ensure_market_status_cachability(self, exchange_manager):
        client_using_cached_markets = ccxt_client_util.ccxt_exchange_class_factory(self.EXCHANGE_NAME)()
        ccxt_client_util.load_markets_from_cache(client_using_cached_markets)
        assert exchange_manager.exchange.connector.client.markets == client_using_cached_markets.markets

    async def get_symbol_prices(self, limit=None, **kwargs):
        async with self.get_exchange_manager() as exchange_manager:
            return await exchange_manager.exchange.get_symbol_prices(self.SYMBOL, self.TIME_FRAME,
                                                                     limit=limit, **kwargs)

    async def get_historical_ohlcv(self, **kwargs) -> list:
        async with self.get_exchange_manager() as exchange_manager:
            start, end = self.get_historical_ohlcv_start_and_end_times()
            start_ms, end_ms = start * 1000, end * 1000
            ohlcvs = []
            async for ohlcv in exchanges_util.get_historical_ohlcv(
                exchange_manager, self.SYMBOL, self.TIME_FRAME, start_ms, end_ms, request_retry_timeout=2, **kwargs
            ):
                ohlcvs.extend(ohlcv)
            return ohlcvs

    def get_historical_ohlcv_start_and_end_times(self):
        start = self.get_time() - (self.get_timeframe_seconds() * self.HISTORICAL_CANDLES_TO_FETCH_COUNT * 2)
        end = self.get_time_after_time_frames(start, self.HISTORICAL_CANDLES_TO_FETCH_COUNT)
        return start, end

    async def get_kline_price(self, **kwargs):
        async with self.get_exchange_manager() as exchange_manager:
            return await exchange_manager.exchange.get_kline_price(self.SYMBOL, self.TIME_FRAME, **kwargs)

    async def get_order_book(self, **kwargs):
        async with self.get_exchange_manager() as exchange_manager:
            return await exchange_manager.exchange.get_order_book(self.SYMBOL, **kwargs)

    async def get_recent_trades(self, limit=50):
        async with self.get_exchange_manager() as exchange_manager:
            return await exchange_manager.exchange.get_recent_trades(self.SYMBOL, limit=limit)

    async def get_price_ticker(self):
        async with self.get_exchange_manager() as exchange_manager:
            return await exchange_manager.exchange.get_price_ticker(self.SYMBOL)

    async def get_all_currencies_price_ticker(self, market_filter=None, **kwargs):
        if market_filter is not None:
            async with self.get_exchange_manager(market_filter=market_filter) as exchange_manager:
                # create 2 exchange manager to force applying market_filter (avoid single test call side effect)
                await exchange_manager.exchange.connector.load_symbol_markets()
        async with self.get_exchange_manager(market_filter=market_filter) as exchange_manager:
            return await exchange_manager.exchange.get_all_currencies_price_ticker(**kwargs)

    def get_market_filter(self):
        def market_filter(market):
            return (
                market[trading_enums.ExchangeConstantsMarketStatusColumns.SYMBOL.value]
                in (self.SYMBOL, self.SYMBOL_2)
            )

        return market_filter

    def get_allowed_time_delta(self):
        return (self.ALLOWED_TIMEFRAMES_WITHOUT_CANDLE + 1) * \
            commons_enums.TimeFramesMinutes[self.TIME_FRAME] * \
            constants.MINUTE_TO_SECONDS * 1.3

    @staticmethod
    def get_time():
        return Exchange.seconds()

    @staticmethod
    def get_ms_time():
        return int(Exchange.milliseconds())

    def get_timeframe_seconds(self):
        return commons_enums.TimeFramesMinutes[self.TIME_FRAME] * constants.MINUTE_TO_SECONDS

    def get_time_after_time_frames(self, start, time_frames_count):
        return start + self.get_timeframe_seconds() * time_frames_count

    def get_timeframe_ms_delta(self, time_frames_count):
        return self.get_ms_time() - (self.get_timeframe_seconds() * time_frames_count * constants.MSECONDS_TO_SECONDS)

    @staticmethod
    def ensure_elements_order(elements, sort_key, reverse=False):
        assert sorted(elements, key=lambda x: x[sort_key], reverse=reverse) == elements

    @staticmethod
    def ensure_unique_elements(elements, key):
        assert len(elements) == len(set(element[key] for element in elements))

    def check_market_status_limits(self, market_status,
                                   normal_price_max=10000, normal_price_min=1e-06,
                                   normal_cost_max=10000, normal_cost_min=1e-06,
                                   low_price_max=1e-07, low_price_min=1e-09,
                                   low_cost_max=1e-03, low_cost_min=1e-06,
                                   expect_invalid_price_limit_values=False,
                                   expect_inferior_or_equal_price_and_cost=False,
                                   enable_price_and_cost_comparison=True,
                                   has_price_limits=True):
        min_price = market_status[Ecmsc.LIMITS.value][Ecmsc.LIMITS_PRICE.value][Ecmsc.LIMITS_PRICE_MIN.value]
        max_price = market_status[Ecmsc.LIMITS.value][Ecmsc.LIMITS_PRICE.value][Ecmsc.LIMITS_PRICE_MAX.value]
        min_cost = market_status[Ecmsc.LIMITS.value][Ecmsc.LIMITS_COST.value][Ecmsc.LIMITS_COST_MIN.value]
        has_min_or_max_price = min_price not in (None, 0) or max_price not in (None, 0)
        if has_min_or_max_price and not has_price_limits:
            raise AssertionError(
                f"Expect no price limit values but min or max price limit is set (min: {min_price} max: {max_price})"
            )
        if not has_min_or_max_price and has_price_limits:
            raise AssertionError(
                f"Expect price limit values but min or max price limit is not set (min: {min_price} max: {max_price})"
            )
        has_price_limit_value = min_price not in (None, 0)
        has_cost_limit_value = min_cost not in (None, 0)
        has_limit_value = has_price_limit_value and has_cost_limit_value
        if (not has_limit_value) and expect_invalid_price_limit_values:
            raise AssertionError("No price and limit value does not mean invalid values")
        if not expect_invalid_price_limit_values and market_status[Ecmsc.SYMBOL.value] == self.SYMBOL_3:
            # if these test are not passing, it means that limits are invalid
            # (limits should be much lower for SYMBOL_3 which is the low price pair, ex: XRP/BTC)
            # => set expect_invalid_price_limit_values to True in call and
            # remove price limit in exchange tentacle market status fixer if this is the case
            if has_price_limit_value:
                assert (not has_price_limit_value) or low_price_max >= min_price >= low_price_min, (
                    f"FALSE: {low_price_max} >= {min_price} >= {low_price_min}"
                )
            assert (not has_cost_limit_value) or low_cost_max >= min_cost >= low_cost_min, (
                    f"FALSE: {low_cost_max} >= {min_cost} >= {min_cost}"
                )
        else:
            if has_price_limit_value:
                assert (not has_price_limit_value) or normal_price_max >= min_price >= normal_price_min
            assert (not has_cost_limit_value) or normal_cost_max >= min_cost >= normal_cost_min
        if has_price_limits and enable_price_and_cost_comparison and has_limit_value:
            # Consistency here is not required by OctoBot. Fix in tentacles if consistency
            # in price/cost comparison becomes required and min_price <= min_cost is false without a good reason
            if expect_inferior_or_equal_price_and_cost:
                assert min_price >= min_cost
            else:
                assert min_price <= min_cost

    @staticmethod
    def check_ticker_typing(ticker, check_open=True, check_high=True, check_low=True,
                            check_close=True, check_base_volume=True, check_timestamp=True):
        if check_open:
            assert isinstance(ticker[Ectc.OPEN.value], (float, int))
        if check_high:
            assert isinstance(ticker[Ectc.HIGH.value], (float, int))
        if check_low:
            assert isinstance(ticker[Ectc.LOW.value], (float, int))
        if check_close:
            assert isinstance(ticker[Ectc.CLOSE.value], (float, int))
        if check_base_volume:
            assert isinstance(ticker[Ectc.BASE_VOLUME.value], (float, int))
        if check_timestamp:
            assert isinstance(ticker[Ectc.TIMESTAMP.value], (float, int))



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import cachetools
import json

import octobot_commons.constants as commons_constants


# To avoid side effects related to a cache refresh at a fix time of the day every day,
# cache should not be refreshed at the same time every day.
# Use 30h and 18min as a period. It could be anything else as long as it doesn't make it so
# that cache ends up refreshed approximately at the same time of the day
_CACHE_TIME = commons_constants.HOURS_TO_SECONDS * 30 + commons_constants.MINUTE_TO_SECONDS * 18
_MARKETS_BY_EXCHANGE = cachetools.TTLCache(maxsize=50, ttl=_CACHE_TIME)


def get_client_key(client) -> str:
    return f"{client.__class__.__name__}:{json.dumps(client.urls.get('api'))}"


def get_exchange_parsed_markets(client_key: str):
    return _MARKETS_BY_EXCHANGE[client_key]


def set_exchange_parsed_markets(client_key: str, markets):
    _MARKETS_BY_EXCHANGE[client_key] = markets



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import sys
import concurrent.futures as futures

import octobot_commons.html_util as html_util

import octobot_commons.thread_util as thread_util
import octobot_trading.constants
import octobot_trading.enums
import octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket


class WebSocketExchange(abstract_websocket.AbstractWebsocketExchange):
    def __init__(self, config, exchange_manager):
        super().__init__(config, exchange_manager)
        self.exchange_manager = exchange_manager
        self.exchange_name = exchange_manager.exchange_name

        self.websocket_connectors = []
        self.websocket_connectors_tasks = []

        self.websocket_connectors_executors = None
        self.websocket_connector = None

        self.pairs = []
        self.time_frames = []

        self.channels = []
        self.handled_feeds = {}

        self.is_websocket_running = False
        self.is_websocket_authenticated = False
        self.is_beyond_feed_exchange_limit = False

        self.restart_task = None

    @classmethod
    def get_exchange_connector_class(cls, exchange_manager):
        raise NotImplementedError("get_exchange_connector_class is not implemented")

    def create_feeds(self):
        raise NotImplementedError("create_feeds is not implemented")

    async def init_websocket(self, time_frames, trader_pairs, tentacles_setup_config):
        self.websocket_connector = self.get_exchange_connector_class(self.exchange_manager)
        self.websocket_connector.update_exchange_feeds(self.exchange_manager)
        self.pairs = trader_pairs
        self.time_frames = time_frames

        if self.pairs:
            if self._has_too_many_feeds_to_handle():
                self.logger.info(
                    f"Disabling websocket: {self.get_connector_feeds_count()} "
                    f"feeds are required and {self.get_connector_max_handled_feeds()} are supported at most."
                )
                self.is_beyond_feed_exchange_limit = True
                return
            # unauthenticated
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.TRADES)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.L1_BOOK)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.L2_BOOK)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.L3_BOOK)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.BOOK_DELTA)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.BOOK_TICKER)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.OPEN_INTEREST)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.MINI_TICKER)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.TICKER)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.FUNDING)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.MARK_PRICE)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.VOLUME)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.LIQUIDATIONS)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.FUTURES_INDEX)

            if self.time_frames:
                await self.add_feed(octobot_trading.enums.WebsocketFeeds.CANDLE)
                await self.add_feed(octobot_trading.enums.WebsocketFeeds.KLINE)

            # authenticated
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.POSITION)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.PORTFOLIO)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.TRADE)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.ORDERS)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.CREATE_ORDER)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.CANCEL_ORDER)
            await self.add_feed(octobot_trading.enums.WebsocketFeeds.LEDGER)

            # ensure feeds are added
            self.create_feeds()
        else:
            self.logger.warning(
                f"{self.exchange_manager.exchange_name.title()}'s websocket has no symbol to feed"
            )

    async def add_feed(self, feed_name):
        if self.is_feed_available(feed_name):
            self.channels.append(feed_name)
            self.handled_feeds[feed_name] = True
            self.logger.debug(f"{self.exchange_manager.exchange_name}'s websocket is handling {feed_name.value}")
        else:
            self.handled_feeds[feed_name] = False
            self.logger.debug(f"{self.exchange_manager.exchange_name}'s websocket is not handling {feed_name.value}")

    def is_feed_available(self, feed):
        try:
            feed_available = self.websocket_connector.get_exchange_feed(feed)
            return feed_available is not octobot_trading.enums.WebsocketFeeds.UNSUPPORTED.value
        except (KeyError, ValueError):
            return False

    def is_feed_requiring_init(self, feed):
        return self.websocket_connector.is_feed_requiring_init(feed)

    @classmethod
    def get_name(cls):
        return cls.__name__

    @classmethod
    def has_name(cls, exchange_manager: object) -> bool:  # pylint: disable=arguments-renamed
        return cls.get_exchange_connector_class(exchange_manager) is not None

    def is_time_frame_related_feed(self, feed):
        return self.websocket_connectors[0].is_time_frame_related_feed(feed)

    def is_time_frame_supported(self, time_frame):
        return all([
            connector.is_time_frame_supported(time_frame)
            for connector in self.websocket_connectors
        ])

    async def start_sockets(self):
        if any(self.handled_feeds.values() and self.websocket_connectors):
            try:
                self.websocket_connectors_executors = futures.ThreadPoolExecutor(
                    max_workers=len(self.websocket_connectors),
                    thread_name_prefix=f"{self.get_name()}-{self.exchange_name}-pool-executor",

                )

                self.websocket_connectors_tasks = [
                    asyncio.get_event_loop().run_in_executor(self.websocket_connectors_executors, websocket.start)
                    for websocket in self.websocket_connectors
                ]

                self.is_websocket_running = True
            except ValueError as e:
                self.logger.error(
                    f"Failed to start websocket on {self.exchange_name} : {html_util.get_html_summary_if_relevant(e)}"
                )

        if self.websocket_connectors and not self.is_websocket_running:
            self.logger.debug(f"{self.exchange_manager.exchange_name.title()}'s "
                              f"websocket is not handling anything, it will not be started.")

    async def wait_sockets(self):
        await asyncio.wait(self.websocket_connectors_tasks)

    def _supports_live_pair_addition(self):
        for websocket in self.websocket_connectors:
            if not websocket.SUPPORTS_LIVE_PAIR_ADDITION:
                return False
        return True

    async def updated_followed_pairs(self):
        for websocket in self.websocket_connectors:
            websocket.update_followed_pairs()

    async def _inner_close_and_restart_sockets(self, debounce_duration):
        # asyncio.sleep to make it easily cancellable to reschedule later calls
        await asyncio.sleep(debounce_duration)
        for websocket in self.websocket_connectors:
            await websocket.reset()

    async def handle_new_pairs(self, debounce_duration=0):
        if self._supports_live_pair_addition():
            await self.updated_followed_pairs()
        else:
            await self._close_and_restart_sockets(debounce_duration=debounce_duration)

    async def _close_and_restart_sockets(self, debounce_duration=0):
        if self.restart_task is not None and not self.restart_task.done():
            self.restart_task.cancel()
        self.restart_task = asyncio.create_task(self._inner_close_and_restart_sockets(debounce_duration))

    async def stop_sockets(self):
        """
        Stops the websocket. Can be restarted
        """
        try:
            for websocket in self.websocket_connectors:
                await websocket.stop()
        except Exception as e:
            self.logger.error(f"Error when stopping sockets : {html_util.get_html_summary_if_relevant(e)}")

    async def close_sockets(self):
        """
        Closes the websocket. Can't be restarted
        """
        try:
            for websocket in self.websocket_connectors:
                await websocket.close()
        except Exception as e:
            self.logger.error(f"Error when closing sockets : {html_util.get_html_summary_if_relevant(e)}")
        for websocket_task in self.websocket_connectors_tasks:
            websocket_task.cancel()
        if sys.version_info.minor >= 9:
            self.websocket_connectors_executors.shutdown(True)
        else:
            thread_util.stop_thread_pool_executor_non_gracefully(self.websocket_connectors_executors)
        self.websocket_connectors_executors = None

    def add_pairs(self, pairs, watching_only=False):
        for websocket in self.websocket_connectors:
            websocket.add_pairs(pairs, watching_only=watching_only)

    def _has_too_many_feeds_to_handle(self) -> bool:
        if self.get_connector_max_handled_feeds() == octobot_trading.constants.NO_DATA_LIMIT:
            return False
        return self.get_connector_feeds_count() >= self.get_connector_max_handled_feeds()

    def get_connector_feeds_count(self):
        return self.websocket_connector.get_feeds_count(self.pairs, self.time_frames)

    def get_connector_max_handled_feeds(self):
        return self.websocket_connector.MAX_HANDLED_FEEDS

    def clear(self):
        super(WebSocketExchange, self).clear()
        for connector in self.websocket_connectors:
            connector.clear()



#  Drakkar-Software OctoBot-Private-Tentacles
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal
import time
import typing

import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.enums as common_enums
import octobot_commons.constants as commons_constants
import octobot_commons.logging as logging
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.api as trading_api
import octobot_trading.exchanges as exchanges
import octobot_trading.exchange_data as trading_exchange_data
import octobot_tentacles_manager.api as tentacles_manager_api

import octobot_trading.personal_data as personal_data

import octobot_trading.util.test_tools.exchange_data as exchange_data_import


BASE_TIMEOUT = 10


async def create_test_exchange_manager(
        config: dict,
        exchange_name: str,
        is_spot_only: bool = True,
        is_margin: bool = False,
        is_future: bool = False,
        rest_only: bool = False,
        is_real: bool = True,
        is_sandboxed: bool = False,
        ignore_exchange_config: bool = True) -> exchanges.ExchangeManager:
    if ignore_exchange_config:
        # enable exchange name in config
        config[commons_constants.CONFIG_EXCHANGES][exchange_name] = {commons_constants.CONFIG_ENABLED_OPTION: True}

    builder = exchanges.create_exchange_builder_instance(config, exchange_name)
    builder.disable_trading_mode()
    builder.use_tentacles_setup_config(tentacles_manager_api.get_tentacles_setup_config())
    if is_spot_only:
        builder.is_spot_only()
    if is_margin:
        builder.is_margin()
    if is_future:
        builder.is_future()
    if rest_only:
        builder.is_rest_only()
    if is_sandboxed:
        builder.is_sandboxed(is_sandboxed)
    if is_real:
        builder.is_real()
    else:
        builder.is_simulated()
    await builder.build()
    return builder.exchange_manager


async def stop_test_exchange_manager(exchange_manager_instance: exchanges.ExchangeManager):
    trading_api.cancel_ccxt_throttle_task()
    await exchange_manager_instance.stop()
    # let updaters gracefully shutdown
    await asyncio_tools.wait_asyncio_next_cycle()

async def _update_ohlcv(
    exchange_manager, symbol: str, time_frame: str, exchange_data: exchange_data_import.ExchangeData,
    history_size=1, start_time=0, end_time=0, close_price_only=False,
    include_latest_candle=True
):
    parsed_tf = common_enums.TimeFrames(time_frame)
    if start_time == 0:
        ohlcvs = await exchange_manager.exchange.get_symbol_prices(symbol, parsed_tf, limit=history_size)
    else:
        ohlcvs = []
        async for ohlcv in exchanges.get_historical_ohlcv(
            exchange_manager, symbol, parsed_tf, start_time * 1000, (end_time or time.time()) * 1000,
            request_retry_timeout=BASE_TIMEOUT
        ):
            ohlcvs.extend(ohlcv)
    if not include_latest_candle:
        ohlcvs = ohlcvs[:-1]
    details = exchange_data_import.MarketDetails(
        symbol=symbol,
        time_frame=time_frame,
        close=[ohlcv[common_enums.PriceIndexes.IND_PRICE_CLOSE.value] for ohlcv in ohlcvs],
        open=[ohlcv[common_enums.PriceIndexes.IND_PRICE_OPEN.value] for ohlcv in ohlcvs] if not close_price_only else [],
        high=[ohlcv[common_enums.PriceIndexes.IND_PRICE_HIGH.value] for ohlcv in ohlcvs] if not close_price_only else [],
        low=[ohlcv[common_enums.PriceIndexes.IND_PRICE_LOW.value] for ohlcv in ohlcvs] if not close_price_only else [],
        volume=[ohlcv[common_enums.PriceIndexes.IND_PRICE_VOL.value] for ohlcv in ohlcvs] if not close_price_only else [],
        time=[ohlcv[common_enums.PriceIndexes.IND_PRICE_TIME.value] for ohlcv in ohlcvs],
    )
    exchange_data.markets.append(details)


async def add_symbols_details(
    exchange_manager, symbols: list, time_frame: str, exchange_data: exchange_data_import.ExchangeData,
    history_size=1, start_time=0, end_time=0, close_price_only=False,
    include_latest_candle=True, reload_markets=False,
    market_filter: typing.Union[None, typing.Callable[[dict], bool]] = None
) -> exchange_data_import.ExchangeData:

    await ensure_symbol_markets(exchange_manager, reload=reload_markets, market_filter=market_filter)
    if len(symbols) == 1:
        await _update_ohlcv(
            exchange_manager, next(iter(symbols)), time_frame, exchange_data,
            history_size, start_time, end_time, close_price_only, include_latest_candle
        )
    else:
        await asyncio.gather(*(
            _update_ohlcv(
                exchange_manager, symbol, time_frame, exchange_data,
                history_size, start_time, end_time, close_price_only, include_latest_candle
            )
            for symbol in symbols
        ))
    return exchange_data


async def ensure_symbol_markets(
    exchange_manager,
    reload=False,
    market_filter: typing.Union[None, typing.Callable[[dict], bool]] = None
):
    if reload or not exchange_manager.exchange.connector.has_markets():
        await exchange_manager.exchange.connector.load_symbol_markets(reload=reload, market_filter=market_filter)


async def get_portfolio(exchange_manager, as_float=False, clear_empty=True) -> dict:
    balance = await exchange_manager.exchange.get_balance()
    # filter out 0 values
    return {
        asset: {key: float(val) if as_float else val for key, val in values.items()}  # use float for values
        for asset, values in balance.items()
        if not clear_empty or (clear_empty and any(value for value in values.values()))
    }


def _parse_order_dict(
    exchange_manager, order: dict, force_open_or_pending_creation: bool
) -> typing.Optional[personal_data.Order]:
    if not order:
        return None
    try:
        return personal_data.create_order_instance_from_raw(
            exchange_manager.trader, order, force_open_or_pending_creation=force_open_or_pending_creation
        )
    except Exception as err:
        logging.get_logger("_parse_order_dict").exception(
            err,
            True,
            f"Unexpected error when parsing [{exchange_manager.exchange_name}] "
            f"order ({err} {err.__class__.__name__}), order ignored: {order}"
        )
    return None


def _parse_order_into_dict(
    exchange_manager, order: dict, force_open_or_pending_creation: bool, ignore_unsupported_orders: bool
) -> typing.Optional[dict]:
    if (
        ignore_unsupported_orders and
        order[enums.ExchangeConstantsOrderColumns.TYPE.value] == enums.TradeOrderType.UNSUPPORTED.value
    ):
        logging.get_logger("_parse_order_into_dict").warning(
            f"Ignored unsupported [{exchange_manager.exchange_name}] order: {order}"
        )
        return None
    if parsed_order := _parse_order_dict(exchange_manager, order, force_open_or_pending_creation):
        try:
            return parsed_order.to_dict()
        except Exception as err:
            logging.get_logger("_parse_order_dict").exception(
                err,
                True,
                f"Unexpected error when converting [{exchange_manager.exchange_name}] order to dict" 
                f"({err}. {err.__class__.__name__}), order: {order}"
            )
    return None


async def _get_open_orders(exchange_manager, symbol: str, open_orders: list, ignore_unsupported_orders: bool):
    orders = await exchange_manager.exchange.get_open_orders(symbol=symbol)
    for order in orders:
        if order_dict := _parse_order_into_dict(exchange_manager, order, True,  ignore_unsupported_orders):
            open_orders.append(order_dict)


async def get_open_orders(
    exchange_manager,
    exchange_data: exchange_data_import.ExchangeData,
    symbols: list = None,
    ignore_unsupported_orders: bool = True,
) -> list:
    open_orders = []
    symbols = symbols or [market.symbol for market in exchange_data.markets]
    if len(symbols) == 1:
        await _get_open_orders(exchange_manager, next(iter(symbols)), open_orders, ignore_unsupported_orders)
    else:
        await asyncio.gather(*(
            _get_open_orders(exchange_manager, symbol, open_orders, ignore_unsupported_orders) for symbol in symbols
        ))
    return open_orders


async def get_order(
    exchange_manager,
    exchange_order_id: str,
    symbol: str,
) -> typing.Optional[dict]:
    return await exchange_manager.exchange.get_order(exchange_order_id, symbol=symbol)


async def _get_cancelled_orders(exchange_manager, symbol: str, cancelled_orders: list, ignore_unsupported_orders: bool):
    orders = await exchange_manager.exchange.get_cancelled_orders(symbol=symbol)
    for order in orders:
        if order_dict := _parse_order_into_dict(
            exchange_manager, order, False, ignore_unsupported_orders
        ):
            cancelled_orders.append(order_dict)


async def get_cancelled_orders(
    exchange_manager,
    exchange_data: exchange_data_import.ExchangeData,
    symbols: list = None,
    ignore_unsupported_orders: bool = True,
) -> list:
    cancelled_orders = []
    symbols = symbols or [market.symbol for market in exchange_data.markets]
    if len(symbols) == 1:
        await _get_cancelled_orders(exchange_manager, next(iter(symbols)), cancelled_orders, ignore_unsupported_orders)
    else:
        await asyncio.gather(*(
            _get_cancelled_orders(exchange_manager, symbol, cancelled_orders, ignore_unsupported_orders)
            for symbol in symbols
        ))
    return cancelled_orders


async def _get_trades(exchange_manager, symbol: str, trades: list):
    row_trades = await exchange_manager.exchange.get_my_recent_trades(symbol=symbol)
    trades.extend(
        personal_data.create_trade_instance_from_raw(exchange_manager.trader, raw_trade).to_dict()
        for raw_trade in row_trades
    )


async def get_trades(
    exchange_manager,
    exchange_data: exchange_data_import.ExchangeData,
    symbols: list = None
) -> list:
    trades = []
    symbols = symbols or [market.symbol for market in exchange_data.markets]
    if len(symbols) == 1:
        await _get_trades(exchange_manager, next(iter(symbols)), trades)
    else:
        await asyncio.gather(*(
            _get_trades(exchange_manager, symbol, trades)
            for symbol in symbols
        ))
    return trades


async def _create_order(
    exchange_manager,
    order_dict: dict,
    order_creation_timeout: float,
    price_by_symbol: dict[str, float],
) -> typing.Optional[personal_data.Order]:
    symbol = order_dict[enums.ExchangeConstantsOrderColumns.SYMBOL.value]
    side, order_type = personal_data.parse_order_type(order_dict)
    temp_order = _parse_order_dict(exchange_manager, order_dict, True)
    if temp_order is None:
        logging.get_logger("_create_order").error(f"Unexpected: order can't be created.")
        return None
    order_params = exchange_manager.exchange.get_order_additional_params(temp_order)
    created_order = await exchange_manager.exchange.create_order(
        order_type,
        symbol,
        decimal.Decimal(str(order_dict[enums.ExchangeConstantsOrderColumns.AMOUNT.value])),
        price=decimal.Decimal(str(order_dict[enums.ExchangeConstantsOrderColumns.PRICE.value])),
        side=side,
        current_price=price_by_symbol[symbol],
        reduce_only=order_dict[enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value],
        params=order_params
    )
    # is private, to use in tests context only
    order = _parse_order_dict(exchange_manager, created_order, True)
    if order is None:
        logging.get_logger("_create_order").error(f"Unexpected: order hasn't been created.")
        return order
    if order.status is enums.OrderStatus.PENDING_CREATION and order_creation_timeout > 0:
        try:
            return await wait_for_other_status(order, order_creation_timeout)
        except TimeoutError as err:
            logging.get_logger(order.get_logger_name()).error(f"Created order can't be fetched: {err}")
            return None
    return order


async def create_orders(
    exchange_manager,
    orders: list,
    order_creation_timeout: float,
    price_by_symbol: dict[str, float],
) -> list:
    if len(orders) == 1:
        return [
            await _create_order(exchange_manager, next(iter(orders)), order_creation_timeout, price_by_symbol)
        ]
    return await asyncio.gather(*(
        _create_order(exchange_manager, order_dict, order_creation_timeout, price_by_symbol)
        for order_dict in orders
    ))


async def wait_for_other_status(order: personal_data.Order, timeout) -> personal_data.Order:
    t0 = time.time()
    iterations = 0
    origin_status = order.status.value
    while time.time() - t0 < timeout:
        raw_order = await order.exchange_manager.exchange.get_order(order.exchange_order_id, order.symbol)
        iterations += 1
        if raw_order is not None and raw_order[enums.ExchangeConstantsOrderColumns.STATUS.value] != origin_status:
            logging.get_logger(order.get_logger_name()).info(
                f"Order fetched with status different from {origin_status} after {iterations} "
                f"iterations and {round(time.time() - t0)}s"
            )
            if parsed_order := _parse_order_dict(order.exchange_manager, raw_order, False):
                return parsed_order
        if time.time() - t0 + constants.CREATED_ORDER_FORCED_UPDATE_PERIOD >= timeout:
            break
        await asyncio.sleep(constants.CREATED_ORDER_FORCED_UPDATE_PERIOD)
    raise TimeoutError(f"Order was not found with another status than {origin_status} within {timeout} seconds")


async def get_positions(
    exchange_manager,
    exchange_data: exchange_data_import.ExchangeData,
    symbols: list = None
) -> list[dict]:
    symbols = symbols or [market.symbol for market in exchange_data.markets]
    if not symbols:
        # nothing to fetch
        return []
    raw_positions = await exchange_manager.exchange.get_positions(symbols=symbols)
    # initialize relevant contracts first as they might be waited for
    trading_exchange_data.update_contracts_from_positions(exchange_manager, raw_positions)
    dict_positions = [
        personal_data.create_position_instance_from_raw(
            exchange_manager.trader, raw_position
        ).to_dict()
        for raw_position in raw_positions
    ]
    return dict_positions



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses
import typing
import octobot_trading.constants


@dataclasses.dataclass
class ProxyConfig:
    # REST proxy
    http_proxy: typing.Optional[str] = None
    http_proxy_callback: typing.Optional[typing.Callable] = None
    https_proxy: typing.Optional[str] = None
    https_proxy_callback: typing.Optional[typing.Callable] = None
    # Websocket proxy
    socks_proxy : typing.Optional[str] = None
    socks_proxy_callback: typing.Optional[typing.Callable] = None
    # enable trust_env in exchange's aiohttp.ClientSession
    aiohttp_trust_env: bool = octobot_trading.constants.ENABLE_EXCHANGE_HTTP_PROXY_FROM_ENV
    # if set, will be called when exchange stops
    stop_proxy_callback: typing.Optional[typing.Callable] = None
    # if set, returns the last url given to a callback method that return "True", meaning the last url that used a proxy
    get_last_proxied_request_url: typing.Optional[typing.Callable[[], typing.Optional[str]]] = None
    get_proxy_url: typing.Optional[typing.Callable[[], str]] = None
    # the host of this proxy, used to identify proxy connexion errors
    proxy_host: str = "DEFAULT PROXY HOST"
    # if DNS cache should be disabled with this proxy config. Warning: in this case, DNS cache will be
    # completely disabled for this exchange
    disable_dns_cache: bool = False

    @classmethod
    def default_env_var_config(cls):
        return cls(
            http_proxy=octobot_trading.constants.EXCHANGE_HTTP_PROXY_AUTHENTICATED_URL or None,
            https_proxy=octobot_trading.constants.EXCHANGE_HTTPS_PROXY_AUTHENTICATED_URL or None,
            socks_proxy=octobot_trading.constants.EXCHANGE_SOCKS_PROXY_AUTHENTICATED_URL or None,
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal

import async_channel.constants as async_channel_constants

import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.enums as enums
import octobot_trading.constants as constants


class TickerProducer(exchanges_channel.ExchangeChannelProducer):
    async def push(self, symbol, ticker):
        await self.perform(symbol, ticker)
        await self._on_ticker_push(symbol, ticker)

    async def perform(self, symbol, ticker):
        try:
            if self.channel.get_filtered_consumers(symbol=async_channel_constants.CHANNEL_WILDCARD) or \
                    self.channel.get_filtered_consumers(symbol=symbol):  #
                if ticker:  # and price_ticker_is_initialized
                    self.channel.exchange_manager.get_symbol_data(symbol).handle_ticker_update(ticker)
                    await self.send(cryptocurrency=self.channel.exchange_manager.exchange.
                                    get_pair_cryptocurrency(symbol),
                                    symbol=symbol,
                                    ticker=ticker)
        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering update: {e}")

    async def send(self, cryptocurrency, symbol, ticker):
        for consumer in self.channel.get_filtered_consumers(symbol=symbol):
            await consumer.queue.put({
                "exchange": self.channel.exchange_manager.exchange_name,
                "exchange_id": self.channel.exchange_manager.id,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "ticker": ticker
            })

    async def _on_ticker_push(self, symbol, ticker):
        await self._push_mini_ticker(symbol, ticker)
        if self.channel.exchange_manager.is_future:
            await self._push_future_data(symbol, ticker)

    async def _push_mini_ticker(self, pair, ticker):
        """
        Mini ticker
        """
        try:
            if not ticker[enums.ExchangeConstantsTickersColumns.CLOSE.value]:
                # no need to push when there is not even a close price (0 or None)
                return
            await exchanges_channel.get_chan(
                constants.MINI_TICKER_CHANNEL,
                self.channel.exchange_manager.id
            ).get_internal_producer().push(
                pair,
                {
                    enums.ExchangeConstantsMiniTickerColumns.HIGH_PRICE.value:
                        ticker[enums.ExchangeConstantsTickersColumns.HIGH.value],
                    enums.ExchangeConstantsMiniTickerColumns.LOW_PRICE.value:
                        ticker[enums.ExchangeConstantsTickersColumns.LOW.value],
                    enums.ExchangeConstantsMiniTickerColumns.OPEN_PRICE.value:
                        ticker[enums.ExchangeConstantsTickersColumns.OPEN.value],
                    enums.ExchangeConstantsMiniTickerColumns.CLOSE_PRICE.value:
                        ticker[enums.ExchangeConstantsTickersColumns.CLOSE.value],
                    enums.ExchangeConstantsMiniTickerColumns.VOLUME.value:
                        ticker[enums.ExchangeConstantsTickersColumns.BASE_VOLUME.value],
                    enums.ExchangeConstantsMiniTickerColumns.TIMESTAMP.value:
                        ticker[enums.ExchangeConstantsTickersColumns.TIMESTAMP.value]
                }
            )
        except Exception as e:
            self.logger.error(f"Failed to parse mini ticker : {e}")

    async def _push_future_data(self, symbol: str, ticker: dict):
        if self.channel.exchange_manager.exchange.MARK_PRICE_IN_TICKER:
            await self._push_mark_price(symbol, ticker)

        if self.channel.exchange_manager.exchange.FUNDING_IN_TICKER:
            await self._push_funding_rate(symbol, ticker)

    async def _push_mark_price(self, symbol: str, ticker: dict):
        try:
            ticker = self.channel.exchange_manager.exchange.parse_mark_price(ticker, from_ticker=True)
            await exchanges_channel.get_chan(constants.MARK_PRICE_CHANNEL,
                                             self.channel.exchange_manager.id).get_internal_producer().push(
                symbol,
                decimal.Decimal(str(ticker[enums.ExchangeConstantsMarkPriceColumns.MARK_PRICE.value])),
                mark_price_source=enums.MarkPriceSources.TICKER_CLOSE_PRICE.value
            )
        except Exception as e:
            self.logger.exception(e, True, f"Fail to update mark price from ticker : {e}")

    async def _push_funding_rate(self, symbol: str, ticker: dict):
        try:
            funding_from_ticker = self.channel.exchange_manager.exchange.parse_funding(ticker, from_ticker=True)
            if not funding_from_ticker:
                # not enough info, don't push possibly false data
                return
            await exchanges_channel.get_chan(constants.FUNDING_CHANNEL, self.channel.exchange_manager.id)\
                .get_internal_producer().push(
                    symbol,
                    funding_from_ticker[enums.ExchangeConstantsFundingColumns.FUNDING_RATE.value],
                    funding_from_ticker[enums.ExchangeConstantsFundingColumns.PREDICTED_FUNDING_RATE.value],
                    funding_from_ticker[enums.ExchangeConstantsFundingColumns.NEXT_FUNDING_TIME.value],
                    funding_from_ticker[enums.ExchangeConstantsFundingColumns.LAST_FUNDING_TIME.value],
            )
        except Exception as e:
            self.logger.exception(e, True, f"Fail to update funding rate from ticker : {e}")


class TickerChannel(exchanges_channel.ExchangeChannel):
    PRODUCER_CLASS = TickerProducer
    CONSUMER_CLASS = exchanges_channel.ExchangeChannelConsumer


class MiniTickerProducer(exchanges_channel.ExchangeChannelProducer):
    async def push(self, symbol, mini_ticker):
        await self.perform(symbol, mini_ticker)

    async def perform(self, symbol, mini_ticker):
        try:
            if self.channel.get_filtered_consumers(symbol=async_channel_constants.CHANNEL_WILDCARD) or \
                    self.channel.get_filtered_consumers(symbol=symbol):
                if mini_ticker:
                    self.channel.exchange_manager.get_symbol_data(symbol).handle_mini_ticker_update(mini_ticker)
                    await self.send(cryptocurrency=self.channel.exchange_manager.exchange.
                                    get_pair_cryptocurrency(symbol),
                                    symbol=symbol,
                                    mini_ticker=mini_ticker)
        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering update: {e}")

    async def send(self, cryptocurrency, symbol, mini_ticker):
        for consumer in self.channel.get_filtered_consumers(symbol=symbol):
            await consumer.queue.put({
                "exchange": self.channel.exchange_manager.exchange_name,
                "exchange_id": self.channel.exchange_manager.id,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "mini_ticker": mini_ticker
            })


class MiniTickerChannel(exchanges_channel.ExchangeChannel):
    PRODUCER_CLASS = MiniTickerProducer
    CONSUMER_CLASS = exchanges_channel.ExchangeChannelConsumer
